@@ -2,11 +2,11 @@
 /*                                                                          */
 /*  The FreeType project -- a free and portable quality TrueType renderer.  */
 /*                                                                          */
-/*  Copyright 1996-1999 by                                                  */
+/*  Copyright 1996-2002, 2003, 2004 by                                      */
 /*  D. Turner, R.Wilhelm, and W. Lemberg                                    */
 /*                                                                          */
 /*                                                                          */
-/*  FTString.c - simple text string display                                 */
+/*  ftstring.c - simple text string display                                 */
 /*                                                                          */
 /****************************************************************************/
 
@@ -37,29 +37,34 @@
 
   static char*  Text = (char *)"The quick brown fox jumps over the lazy dog";
 
-  static FT_Library  library;       /* the FreeType library            */
-  static FT_Face     face;          /* the font face                   */
-  static FT_Error    error;         /* error returned by FreeType ?    */
+  static FT_Library   library;      /* the FreeType library            */
+  static FT_Face      face;         /* the font face                   */
+  static FT_Error     error;        /* error returned by FreeType?     */
 
-  static grSurface*  surface;       /* current display surface         */
-  static grBitmap    bit;           /* current display bitmap          */
+  static FT_Encoding  encoding = FT_ENCODING_UNICODE;
+
+  static grSurface*   surface;      /* current display surface         */
+  static grBitmap     bit;          /* current display bitmap          */
 
   static int  ptsize;               /* current point size */
   static int  Num;
   static int  Rotation = 0;
   static int  Fail;
 
-  static int  hinted    = 1;        /* is glyph hinting active ?    */
-  static int  antialias = 1;        /* is anti-aliasing active ?    */
-  static int  use_sbits = 0;        /* do we use embedded bitmaps ? */
-  static int  kerning   = 1;
+  static int  hinted    = 1;        /* is glyph hinting active?        */
+  static int  antialias = 1;        /* is anti-aliasing active?        */
+  static int  use_sbits = 0;        /* do we use embedded bitmaps?     */
+  static int  kerning   = 2;        /* 0: no kerning;                  */
+                                    /* 1: `kern' values                */
+                                    /* 2: `kern' + side bearing errors */
+  static int  autohint  = 0;        /* is forced hinting active?       */
+  static int  use_gamma = 0;
 
   static int  res = 72;             /* default resolution in dpi */
 
   static grColor  fore_color = { 255 };
 
   static int  graph_init  = 0;
-  static int  render_mode = 1;
 
   static FT_Matrix  trans_matrix;
   static int        transform = 0;
@@ -121,6 +126,25 @@
   }
 
 
+  static unsigned long
+  make_tag( char  *s )
+  {
+    int            i;
+    unsigned long  l = 0;
+
+
+    for ( i = 0; i < 4; i++ )
+    {
+      if ( !s[i] )
+        break;
+      l <<= 8;
+      l  += (unsigned long)s[i];
+    }
+
+    return l;
+  }
+
+
 /****************************************************************************/
 /****************************************************************************/
 /****************************************************************************/
@@ -139,9 +163,60 @@
   /*  up a string of text.                                                 */
   /*                                                                       */
   /*                                                                       */
-  static TGlyph  glyphs[ MAX_GLYPHS ];
+  static TGlyph  glyphs[MAX_GLYPHS];
   static int     num_glyphs;
 
+  /*************************************************************************/
+  /*                                                                       */
+  /* Gamma correction.                                                     */
+  /*                                                                       */
+  /*                                                                       */
+  static double  gamma_value = 2.0f;
+  static FT_Byte gamma_ramp[256];
+
+  static void
+  init_gamma( void )
+  {
+    int i;
+    double gamma_inv = 1.0f / gamma_value;
+
+    for (i = 0; i < 256; i++)
+      gamma_ramp[i] = (FT_Byte)( pow((double)i / 255.0f, gamma_inv) * 255.0f );
+  }
+
+  static void
+  apply_gamma( grBitmap* bmp )
+  {
+    int       i, j;
+    FT_Byte*  buffer = (FT_Byte*) bmp->buffer;
+
+    for (i = 0; i < bmp->rows; i++)
+    {
+      for (j = 0; j < bmp->width; j++)
+        buffer[j] = (FT_Byte) gamma_ramp[buffer[j]];
+
+      buffer += bmp->pitch;
+    }
+  }
+
+  static void
+  draw_gamma_ramp( void )
+  {
+    int   i, x, y;
+    long  pitch = bit.pitch;
+    long  start = 0;
+
+    if ( pitch < 0 )
+      start = -pitch*(bit.rows-1);
+
+    x = (bit.width - 256) / 2;
+    y = (bit.rows + 256) / 2;
+    for (i = 0; i < 256; i++, x++)
+    {
+      bit.buffer[start + pitch*(y - gamma_ramp[i]) + x ] = 80;
+    }
+  }
+
 
   /*************************************************************************/
   /*                                                                       */
@@ -206,21 +281,19 @@
     PGlyph     glyph = glyphs;
     int        n;
     FT_Vector  origin;
-    FT_Pos     origin_x = 0;
+    FT_Pos     origin_x       = 0;
     FT_UInt    load_flags;
-    FT_UInt    num_grays;
-    FT_UInt    prev_index = 0;
+    FT_UInt    prev_index     = 0;
+    FT_Pos     prev_rsb_delta = 0;
 
 
-    load_flags = FT_LOAD_DEFAULT;
+    load_flags = FT_LOAD_DEFAULT | FT_LOAD_IGNORE_GLOBAL_ADVANCE_WIDTH;
     if ( !hinted )
       load_flags |= FT_LOAD_NO_HINTING;
     if ( !use_sbits )
       load_flags |= FT_LOAD_NO_BITMAP;
-
-    num_grays = 256;
-    if ( !antialias )
-      num_grays = 0;
+    if ( autohint )
+      load_flags |= FT_LOAD_FORCE_AUTOHINT;
 
     for ( n = 0; n < num_glyphs; n++, glyph++ )
     {
@@ -233,7 +306,7 @@
 
 
           FT_Get_Kerning( face, prev_index, glyph->glyph_index,
-                          hinted ? ft_kerning_default : ft_kerning_unfitted,
+                          hinted ? FT_KERNING_DEFAULT : FT_KERNING_UNFITTED,
                           &kern );
 
           origin_x += kern.x;
@@ -241,32 +314,42 @@
         prev_index = glyph->glyph_index;
       }
 
-      origin.x = origin_x;
-      origin.y = 0;
-
       /* clear existing image if there is one */
       if ( glyph->image )
         FT_Done_Glyph( glyph->image );
 
-      /* load the glyph image (in its native format); */
-      /* for now, we take a monochrome glyph bitmap   */
+      /* load the glyph image (in its native format) */
       error = FT_Load_Glyph( face, glyph->glyph_index, load_flags );
       if ( error )
         continue;
 
-      error = FT_Get_Glyph ( face->glyph, &glyph->image );
+      if ( n && kerning > 1 )
+      {
+        if ( prev_rsb_delta - face->glyph->lsb_delta >= 32 )
+          origin_x -= 64;
+
+        else if ( prev_rsb_delta - face->glyph->lsb_delta < -32 )
+          origin_x += 64;
+      }
+
+      origin.x = origin_x;
+      origin.y = 0;
+
+      prev_rsb_delta = face->glyph->rsb_delta;
+
+      error = FT_Get_Glyph( face->glyph, &glyph->image );
       if ( error )
         continue;
 
       glyph->pos = origin;
 
-      origin_x  += face->glyph->advance.x;
+      origin_x += face->glyph->advance.x;
     }
 
     string_center.x = ( origin_x / 2 ) & -64;
     string_center.y = 0;
 
-    if ( transform )
+    if ( transform && FT_IS_SCALABLE( face ) )
       FT_Vector_Transform( &string_center, &trans_matrix );
   }
 
@@ -305,10 +388,12 @@
 
       /* transform it */
       vec = glyph->pos;
-      FT_Vector_Transform( &vec, &trans_matrix );
+      if ( FT_IS_SCALABLE( face ) )
+        FT_Vector_Transform( &vec, &trans_matrix );
       vec.x += delta.x;
       vec.y += delta.y;
-      error = FT_Glyph_Transform( image, &trans_matrix, &vec );
+      if ( FT_IS_SCALABLE( face ) )
+        error = FT_Glyph_Transform( image, &trans_matrix, &vec );
       if ( !error )
       {
         FT_BBox  bbox;
@@ -317,7 +402,17 @@
         /* check bounding box; if it is not within the display surface, */
         /* we don't need to render it                                   */
 
-        FT_Glyph_Get_CBox( image, ft_glyph_bbox_pixels, &bbox );
+        FT_Glyph_Get_CBox( image, FT_GLYPH_BBOX_PIXELS, &bbox );
+        if ( !FT_IS_SCALABLE( face ) )
+        {
+          vec.x = vec.x >> 6;
+          vec.y = vec.y >> 6;
+
+          bbox.xMin += vec.x;
+          bbox.xMax += vec.x;
+          bbox.yMin += vec.y;
+          bbox.yMax += vec.y;
+        }
 
 #if 0
         if ( n == 0 )
@@ -330,10 +425,10 @@
         if ( bbox.xMax > 0         && bbox.yMax > 0        &&
              bbox.xMin < bit.width && bbox.yMin < bit.rows )
         {
-          /* convert to a bitmap - destroy native image */
+          /* convert to a bitmap -- destroy native image */
           error = FT_Glyph_To_Bitmap( &image,
-                                      antialias ? ft_render_mode_normal
-                                                : ft_render_mode_mono,
+                                      antialias ? FT_RENDER_MODE_NORMAL
+                                                : FT_RENDER_MODE_MONO,
                                       0, 1 );
           if ( !error )
           {
@@ -349,7 +444,6 @@
             }
 #endif
 
-
             bit3.rows   = source->rows;
             bit3.width  = source->width;
             bit3.pitch  = source->pitch;
@@ -357,13 +451,15 @@
 
             switch ( source->pixel_mode )
             {
-            case ft_pixel_mode_mono:
+            case FT_PIXEL_MODE_MONO:
               bit3.mode = gr_pixel_mode_mono;
               break;
 
-            case ft_pixel_mode_grays:
+            case FT_PIXEL_MODE_GRAY:
               bit3.mode  = gr_pixel_mode_gray;
               bit3.grays = source->num_grays;
+              if (use_gamma)
+                apply_gamma(&bit3);
               break;
 
             default:
@@ -371,8 +467,16 @@
             }
 
             /* now render the bitmap into the display surface */
-            x_top = bitmap->left;
-            y_top = bit.rows - bitmap->top;
+            if ( FT_IS_SCALABLE( face ) )
+            {
+              x_top = bitmap->left;
+              y_top = bit.rows - bitmap->top;
+            }
+            else
+            {
+              x_top = vec.x - bitmap->left;
+              y_top = vec.y - bitmap->top;
+            }
             grBlitGlyphToBitmap( &bit, &bit3, x_top, y_top, fore_color );
           }
         }
@@ -395,18 +499,69 @@
     const unsigned char*  p     = (const unsigned char*)string;
     PGlyph                glyph = glyphs;
     FT_UInt               glyph_index;
+    unsigned char         in_code;
+    unsigned long         codepoint = 0;
+    int                   expect;
 
 
-    num_glyphs = 0;
+    if ( encoding )
+    {
+      error = FT_Select_Charmap( face, encoding );
+      if ( error )
+        PanicZ( "invalid charmap\n" );
+    }
+
+    num_glyphs = expect = 0;
     while ( *p )
     {
-      glyph_index = FT_Get_Char_Index( face, (FT_ULong)*p );
+      in_code = *p++ ;
+
+      if ( in_code >= 0xC0 )
+      {
+        if ( in_code < 0xE0 )           /*  U+0080 - U+07FF   */
+        {
+          expect = 1;
+          codepoint = in_code & 0x1F;
+        }
+        else if ( in_code < 0xF0 )      /*  U+0800 - U+FFFF   */
+        {
+          expect = 2;
+          codepoint = in_code & 0x0F;
+        }
+        else if ( in_code < 0xF8 )      /* U+10000 - U+10FFFF */
+        {
+          expect = 3;
+          codepoint = in_code & 0x07;
+        }
+        continue;
+      }
+      else if ( in_code >= 0x80 )
+      {
+        --expect;
+
+        if ( expect >= 0 )
+        {
+          codepoint <<= 6;
+          codepoint  += in_code & 0x3F;
+        }
+        if ( expect >  0 )
+          continue;
+
+        expect = 0;
+      }
+      else                              /* ASCII, U+0000 - U+007F */
+        codepoint = in_code;
+
+      if ( encoding )
+        glyph_index = FT_Get_Char_Index( face, codepoint );
+      else
+        glyph_index = codepoint;
+
       glyph->glyph_index = glyph_index;
       glyph++;
       num_glyphs++;
       if ( num_glyphs >= MAX_GLYPHS )
         break;
-      p++;
     }
   }
 
@@ -421,8 +576,8 @@
 
     transform       = ( angle != 0 );
     trans_matrix.xx = cosinus;
-    trans_matrix.xy = -sinus;
     trans_matrix.yx = sinus;
+    trans_matrix.xy = -sinus;
     trans_matrix.yy = cosinus;
   }
 
@@ -458,8 +613,10 @@
     grLn();
     grWriteln("  F1 or ?   : display this help screen" );
     grWriteln("  a         : toggle anti-aliasing" );
+    grWriteln("  f         : toggle forced auto-hinting" );
     grWriteln("  h         : toggle outline hinting" );
     grWriteln("  k         : toggle kerning" );
+    grWriteln("  g         : toggle gamma correction" );
     grLn();
     grWriteln("  Up        : increase pointsize by 1 unit" );
     grWriteln("  Down      : decrease pointsize by 1 unit" );
@@ -471,6 +628,9 @@
     grWriteln("  F7        : big rotate counter-clockwise");
     grWriteln("  F8        : big rotate clockwise");
     grLn();
+    grWriteln("  F9        : decrease gamma by 0.1" );
+    grWriteln("  F10       : increase gamma by 0.1" );
+    grLn();
     grWriteln("press any key to exit this help screen");
 
     grRefreshSurface( surface );
@@ -490,18 +650,25 @@
     case grKEY( 'q' ):
       return 0;
 
+    case grKEY( 'f' ):
+      autohint = !autohint;
+      new_header = autohint ? (char *)"forced auto-hinting is now on"
+                            : (char *)"forced auto-hinting is now off";
+      return 1;
+
     case grKEY( 'k' ):
-      kerning = !kerning;
-      new_header = kerning
-                     ? (char *)"kerning is now active"
-                     : (char *)"kerning is now ignored";
+      kerning = ( kerning + 1 ) % 3;
+      new_header =
+        kerning == 2
+          ? (char *)"kerning and side bearing correction is now active"
+          : kerning == 1 ? (char *)"kerning is now active"
+                         : (char *)"kerning is now ignored";
       return 1;
 
     case grKEY( 'a' ):
       antialias  = !antialias;
-      new_header = antialias
-                     ? (char *)"anti-aliasing is now on"
-                     : (char *)"anti-aliasing is now off";
+      new_header = antialias ? (char *)"anti-aliasing is now on"
+                             : (char *)"anti-aliasing is now off";
       return 1;
 
     case grKEY( 'b' ):
@@ -521,17 +688,15 @@
 
     case grKEY( 'h' ):
       hinted     = !hinted;
-      new_header = hinted
-                     ? (char *)"glyph hinting is now active"
-                     : (char *)"glyph hinting is now ignored";
+      new_header = hinted ? (char *)"glyph hinting is now active"
+                          : (char *)"glyph hinting is now ignored";
       break;
 
-    case grKEY( ' ' ):
-      render_mode ^= 1;
-      new_header   = render_mode
-                       ? (char *)"rendering all glyphs in font"
-                       : (char *)"rendering test text string";
-      break;
+    case grKEY( 'g' ):
+      use_gamma = !use_gamma;
+      new_header = use_gamma ? (char *)"gamma correction is now on"
+                             : (char *)"gamma correction is now off";
+      return 1;
 
     case grKeyF1:
     case grKEY( '?' ):
@@ -554,6 +719,9 @@
     case grKeyRight: i =   1; goto Do_Rotate;
     case grKeyF7:    i = -10; goto Do_Rotate;
     case grKeyF8:    i =  10; goto Do_Rotate;
+
+    case grKeyF9 : if (gamma_value > 0.1f) gamma_value -= 0.1f; goto Do_Gamma_Set;
+    case grKeyF10:                         gamma_value += 0.1f; goto Do_Gamma_Set;
     default:
       ;
     }
@@ -563,6 +731,16 @@
     Rotation = ( Rotation + i ) & 127;
     return 1;
 
+  Do_Gamma_Set:
+  {
+    static char header_buffer[64];
+
+    sprintf(header_buffer, "gamma is %.1f", gamma_value);
+    new_header = header_buffer;
+    init_gamma();
+    return 1;
+  }
+
   Do_Scale:
     ptsize += i;
     if ( ptsize < 1 )         ptsize = 1;
@@ -600,6 +778,7 @@
     fprintf( stderr,  "Usage: %s [options below] ppem fontname[.ttf|.ttc] ...\n",
              execname );
     fprintf( stderr,  "\n" );
+    fprintf( stderr,  "  -e enc      specify encoding tag (default: unic)\n" );
     fprintf( stderr,  "  -r R        use resolution R dpi (default: 72 dpi)\n" );
     fprintf( stderr,  "  -m message  message to display\n" );
     fprintf( stderr,  "\n" );
@@ -612,14 +791,15 @@
   main( int     argc,
         char**  argv )
   {
-    int    i, old_ptsize, orig_ptsize, file;
-    int    first_glyph = 0;
-    int    XisSetup = 0;
-    char   filename[128 + 4];
-    char   alt_filename[128 + 4];
-    char*  execname;
-    int    option;
-    int    file_loaded;
+    int     i, old_ptsize, orig_ptsize, file;
+    int     first_glyph = 0;
+    int     XisSetup = 0;
+    char    filename[128 + 4];
+    char    alt_filename[128 + 4];
+    char*   execname;
+    int     option;
+    int     file_loaded;
+    PGlyph  glyph;
 
     grEvent   event;
 
@@ -628,13 +808,17 @@
 
     while ( 1 )
     {
-      option = getopt( argc, argv, "m:r:" );
+      option = getopt( argc, argv, "e:m:r:" );
 
       if ( option == -1 )
         break;
 
       switch ( option )
       {
+      case 'e':
+        encoding = (FT_Encoding)make_tag( optarg );
+        break;
+
       case 'r':
         res = atoi( optarg );
         if ( res < 1 )
@@ -667,7 +851,7 @@
     /* Initialize engine */
     error = FT_Init_FreeType( &library );
     if ( error )
-      PanicZ( "Could not initialise FreeType library" );
+      PanicZ( "Could not initialize FreeType library" );
 
   NewFile:
     ptsize      = orig_ptsize;
@@ -709,6 +893,8 @@
       goto Display_Font;
 
   Success:
+    init_gamma();
+
     /* prepare the text to be rendered */
     prepare_text( (unsigned char*)Text );
 
@@ -751,6 +937,8 @@
         {
           reset_transform();
           layout_glyphs();
+          if (use_gamma)
+            draw_gamma_ramp();
           render_string( bit.width/2, bit.rows/2 );
         }
 
@@ -813,6 +1001,14 @@
     }
 
   Fin:
+    glyph = glyphs;
+    for ( i = 0; i < MAX_GLYPHS; i++, glyph++ )
+      if ( glyph->image )
+        FT_Done_Glyph( glyph->image );
+
+    FT_Done_Face    ( face    );
+    FT_Done_FreeType( library );
+
 #if 0
     grDoneSurface( surface );
     grDone();
@@ -822,7 +1018,7 @@
 
     exit( 0 );      /* for safety reasons */
     return 0;       /* never reached */
-}
+  }
 
 
 /* End */
